﻿-- Author      : RisM
-- Create Date : 6/28/2009 4:04:27 PM

SpeakinSpell = LibStub("AceAddon-3.0"):GetAddon("SpeakinSpell")
local L	= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell", false)

-------------------------------------------------------------------------------
-- RANDOM SPEECHES - THE CORE FUNCTIONALITY
-------------------------------------------------------------------------------


function SpeakinSpell:OnEventDetected( DetectedEventStub )
	local funcname = "OnEventDetected"

	-- fill in the stub
	local DetectedEvent = self:CreateDetectedEvent( DetectedEventStub )
	
	local rankedkey, ranklesskey = self:MakeEventTableKeys( DetectedEvent )
	
	local SpeakForRankless = true
	
	--NOTE: it looks like I want to make a subroutine here and call it twice
	--		but the logic is actually slightly different between the ranked and rankless cases
	
	-- look for an event table entry for the ranked version
	if rankedkey then
		local de = self:CopyTable( DetectedEvent )
		de.key = rankedkey
		de.EventTableEntry = SpeakinSpellSavedData.EventTable[ de.key ]
		
		if SpeakinSpellSavedData.ShowSpellEvents then
			if de.rank ~= "" then
				self:Print( self:ProcessSubstitutionVars(L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink.." ("..de.rank..")", de) ) --TODOLATER: add <target> and consider <displaylink> to encapsulate "L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink"
			else
				self:Print( self:ProcessSubstitutionVars(L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink, de) ) --TODOLATER: add <target> and consider <displaylink> to encapsulate "L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink"
			end
		end

		if de.EventTableEntry then
			SpeakForRankless = false
			if self:AllowSpeakForSpell( de ) then
				self:SpeakForSpell( de )
			end
		else
			-- record a new event for this ranked key
			self:DebugMsg(funcname, "no EventTableEntry for key:"..tostring(de.key))
			self:RecordNewEvent( de )
			
			-- if this is a new macro, then open the Create New interface to show it
			-- if that is the case, we need to make sure macros can be shown by the selected filter
			if de.type == "MACRO" then
				if not self:MatchesFilter(DetectedEvent,false) then
					self:SetFilter("MACRO","")
				end
				-- make this new item the selection
				self.OptionsGUIStates.CreateNew.SelectedEventKey = de.key
				-- /ss create
				self:ShowCreateNew()
			end
			
		end
	end
	
	-- check for a rankless version, but don't speak for both
	if ranklesskey then
		local de = self:CopyTable( DetectedEvent )
		de.key = ranklesskey
		de.EventTableEntry = SpeakinSpellSavedData.EventTable[ de.key ]

-- don't need these redundant notifications	- we always get a ranked one first	
--		if SpeakinSpellSavedData.ShowSpellEvents then
--			self:Print( self:ProcessSubstitutionVars(L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink.." "..L["(Any Rank)"], de) ) --TODOLATER add <target> and consider <displaylink> to encapsulate "L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink"
--		end
		
		if SpeakinSpellSavedData.EventTable[ de.key ] then
			if SpeakForRankless then
				if self:AllowSpeakForSpell( de ) then
					self:SpeakForSpell( de )
				end
			end
		else
			-- record a new RANKLESS event for this rankless key
			de.rank = "" -- to make the new event rankless -- NOTE: don't erase this before speaking, to maintain validity of <spellrank> variable
			de.DisplayName = self:FormatDisplayName( de ) -- update display name to show "any rank"
			self:DebugMsg(funcname, "no EventTableEntry for key:"..tostring(de.key))
			self:RecordNewEvent(de)
		end
	end

end



function SpeakinSpell:OnRecursiveEventDetected( text, DetectedEventStub, ParentDetectedEvent )
	local funcname = "OnEventDetected"

	-- fill in the stub
	local DetectedEvent = self:CreateDetectedEvent( DetectedEventStub )
	-- note: recursive events only have a ranked key
	local key, _ = self:MakeEventTableKeys( DetectedEvent )
	
	if not key then
		self:DebugMsg(funcname,"ERROR: undefined key")
		return
	end
	self:DebugMsg(funcname,"key:"..key)

	-- look for an event table entry for the ranked version
	local de = self:CopyTable( DetectedEvent )
	de.key = key
	de.EventTableEntry = SpeakinSpellSavedData.EventTable[ de.key ]
	
	if SpeakinSpellSavedData.ShowSpellEvents then
		self:Print( self:ProcessSubstitutionVars(L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink, de) ) --TODOLATER: add <target> and consider <displaylink> to encapsulate "L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink"
	end

	if not de.EventTableEntry then
		self:Print("ERROR: no event settings for "..text)
		return
	end

	if self:AllowSpeakForSpell( de ) then
		self.RuntimeData.ParentDetectedEvent = ParentDetectedEvent
		self:SpeakForSpell( de )
		self.RuntimeData.ParentDetectedEvent = nil
	end
	
end



function SpeakinSpell:SelectRandomMessage(DetectedEvent)
	assert(DetectedEvent.EventTableEntry and DetectedEvent.EventTableEntry.Messages) -- this function should not be called unless we have a valid table for this spell
	
	-- select a random message from the user-specific table
	
	local sizelist = #(DetectedEvent.EventTableEntry.Messages)
	local tempnum = math.random(1, sizelist);
	
	msg = DetectedEvent.EventTableEntry.Messages[tempnum]
	assert(msg)
	assert(type(msg) == "string")
	
	-- avoid repeating the same message we used last time
	-- unless there's only one in the list
	local History = self.RuntimeData.AnnouncementHistory[DetectedEvent.key]
	if History then
		-- we've announced this event before
		-- make sure we select a different speech than last time, if applicable
		while (sizelist > 1) and (msg == History.LastMessage) do
			tempnum = math.random(1, sizelist);
			msg = DetectedEvent.EventTableEntry.Messages[tempnum]
		end
	end
	
	return msg
end


function SpeakinSpell:GetChatChannelForSpell(EventTableEntry)
	local funcname = "GetChatChannelForSpell"
	
	-- NOTE: the EventTableEntry.Channels table stores channel names that are compatible with SendChatMessage
	--	see also CurrentMessagesGUI_OnChannelSelect
	if (IsActiveBattlefieldArena()) then
		return EventTableEntry.Channels.Arena
	elseif (GetNumBattlefieldScores() > 0) then
		return EventTableEntry.Channels.BG
	elseif (self:IsInWintergraspBattle()) then --only applies to active battles
		-- NOTE: raid warnings not enabled for WG battles
		return EventTableEntry.Channels.WG
	elseif (GetNumRaidMembers() > 0) then
		-- override Raid with Leader if applicable
		if IsRaidLeader() then
			return EventTableEntry.Channels.RaidLeader
		elseif IsRaidOfficer() then
			return EventTableEntry.Channels.RaidOfficer
		else
			if EventTableEntry.Channels.Raid == "RAID_WARNING" then
				-- if the player wants to speak in the RAID_WARNING channel, 
				-- but doesn't have permission (not the leader or assist)
				-- then fall back to the RAID channel instead, so we say something at all
				-- instead of triggering an error message in the chat
				self:DebugMsg(funcname, "You don't have permission to say Raid Warnings right now - using RAID chat instead")
				return "RAID"
			end
		end
		-- else we can use the specified Raid channel
		return EventTableEntry.Channels.Raid
	elseif (GetNumPartyMembers() > 0) then
		-- override Party with Leader if applicable
		if IsPartyLeader() then
			return EventTableEntry.Channels.PartyLeader
		else
			if EventTableEntry.Channels.Party == "RAID_WARNING" then
				self:DebugMsg(funcname, "You don't have permission to say Raid Warnings right now - using PARTY chat instead")
				return "PARTY"
			else
				return EventTableEntry.Channels.Party
			end
		end
	else -- solo
		return EventTableEntry.Channels.Solo
	end
end


function SpeakinSpell:GetLanguageList()
	--TODO: use this function somewhere
	--		this is just some sample code captured from wowwiki while I'm looking at it
	
	--Create and return a list of RP languages that are known to the player
	Langs = {}
	for i=1, GetNumLanguages() do
		tinsert(Langs, GetLanguageByIndex(i))
	end
	return Langs
end


function SpeakinSpell:InsertRandomFaction(name)
	if (name ~= nil) and (name ~= "") then
		tinsert(SpeakinSpellSavedData.RandomSubs["randomfaction"], name)
	end
end


function SpeakinSpell:AddRandomFactions()
	-- add the player's guild name (if any)
	-- guildName, guildRankName, guildRankIndex = GetGuildInfo(unit);
	local guildName = GetGuildInfo( UnitName("player") )
	self:InsertRandomFaction(guildName)
	
	-- add the player's arena teams (if any)
	-- teamName, teamSize, teamRating, weekPlayed, weekWins, seasonPlayed, seasonWins, playerPlayed, seasonPlayerPlayed, teamRank, playerRating = GetArenaTeam(id);
	local teamName2 = GetArenaTeam(1);
	local teamName3 = GetArenaTeam(2);
	local teamName5 = GetArenaTeam(5);

	self:InsertRandomFaction(teamName2)
	self:InsertRandomFaction(teamName3)
	self:InsertRandomFaction(teamName5)
end


-- returns a random entry from the L.RANDOM[something] table
-- for example L.RANDOM["randomfaction"] is a list of faction names
-- changed to be customizable in saved data SpeakinSpellSavedData.RandomSubs = L.RANDOM
function SpeakinSpell:GetRandomSubText(key)
	local t = SpeakinSpellSavedData.RandomSubs[key]
	local sizelist = #(t)
	local tempnum = math.random(1, sizelist);
	return t[tempnum]
end


-- Stonarius' or Meneldill's with the proper apostrophe
function SpeakinSpell:MakePossessive(name)
	if self:EndsWith(tostring(name), "s") then
		return tostring(name).."\'" -- Stonarius'
	else
		return tostring(name).."\'s"-- Meneldill's
	end
end


function SpeakinSpell:MakeUnposessive(names)
	if self:EndsWith(tostring(names), "'s") then
		local name = string.sub(names, 0, string.len(names)-2)
		return name
	elseif self:EndsWith(tostring(names), "'") then
		local name = string.sub(names, 0, string.len(names)-1)
		return name
	else
		return names
	end
end


function SpeakinSpell:AddRandomSubstitutions(Substitutions)
	for var,table in pairs(SpeakinSpellSavedData.RandomSubs) do
		--Substitutions["randomfaction"] = self:GetRandomSubText("randomfaction")
		Substitutions[var] = self:GetRandomSubText(var)
		-- possessive form
		Substitutions[ self:MakePossessive(var) ] = self:MakePossessive( self:GetRandomSubText(var) )
	end
end



function SpeakinSpell:BuildSubstitutionTable(DetectedEvent)
	-- NOTE: also support the english variable names if running in a non-english locale
	--		this is meant to help support the fact that documentation in-game is lacking for this variable list
	--		it's also standard for non-English clients to support English commands

	local Substitutions = {		
		--global info
		-- NOTES:
		--		GetRealZoneText is subtly different from GetZoneText re: instance names
		--		see http://www.wowwiki.com/API_GetZoneText
		--		we want to use GetRealZoneText to use real instance names instead of subzone names
		--		<subzone> will use the minimap text for instances
		--		GetSubZoneText returns an empty string if you're not in a sub zone
		--		GetMinimapZoneText returns the current subzone text shown on the minimap
		--		the minimap always shows something = the greater region name if not in a subzone
		--		so this is guaranteed to return an appropriate value
		["selected"] = UnitName("target"),
		["realm"] = GetRealmName(),
		["zone"] = GetRealZoneText(),
		["subzone"] = GetMinimapZoneText(),
		["playertitle"] = GetTitleName( GetCurrentTitle() ),
		["playerfulltitle"] = self:GetPlayerFullTitleName(),
		["guild"] = self:GetGuildName(),

		-- Stonarius' or Meneldill's with the proper apostrophe
		["selected's"] = self:MakePossessive( UnitName("target") ),
		["realm's"] = self:MakePossessive( GetRealmName() ),
		["zone's"] = self:MakePossessive( GetRealZoneText() ),
		["subzone's"] = self:MakePossessive( GetMinimapZoneText() ),
		["guild's"] = self:MakePossessive( self:GetGuildName() ),
		["playerfulltitle's"] = self:MakePossessive( self:GetPlayerFullTitleName() ),

		["home"] = GetBindLocation(),
		["home's"] = self:MakePossessive( GetBindLocation() ),

		-- auto-substitutions like MS Word would use
		["tm"] = "™",	-- Trademark symbol
		["r"] = "®",	-- Registered Trademark symbol
		["c"] = "©",	-- Copyright (C)
	}
	
	-- use parent detected event info for recursive calls, if available
	local de = ""
	if self.RuntimeData.ParentDetectedEvent then
		self:DebugMsg(nil,"using parent detected event info for substitutions in recursive call")
		de = self.RuntimeData.ParentDetectedEvent
	else
		de = DetectedEvent
	end

	-- add event-specific info
	for var,value in pairs(de) do
		local key = string.lower(var)
		if (type(value) == "string") and not Substitutions[key] then --make sure we don't accidentally create unprintable garbage
			Substitutions[key] = value 
		end
	end
	
	-- event info that may be renamed
	local Renames = {
		["spellname"] = de.name, --TODO: rename to match, requires saved data change
		["spellrank"] = de.rank, --TODO: rename to match, requires saved data change
		--["spelllink"] = de.SpellLink, -- NOTE: this case change was done above
	}
	for var,value in pairs(Renames) do
		if not Substitutions[var] then
			Substitutions[var] = value
		end
	end
	
	-- if the event failed to specify anything,
	-- make sure a default value exists
	local DefaultTarget = self:GetDefaultTarget(false)
	local DefaultEventInfo = {
		["caster"] = UnitName("player"), -- the caster's name
		["target"] = DefaultTarget, -- the target's name
		["targetclass"] = UnitClass(DefaultTarget), -- the target's name
		["targetrace"] = UnitRace(DefaultTarget), -- the target's name
		["spellname"] = "Unspecified Event Name", --TODO: localize
		["spellrank"] = "Rank 1",
		["spelllink"] = "[Unspecified Event Name]",
	}
	for var,value in pairs(DefaultEventInfo) do
		if not Substitutions[var] then
			Substitutions[var] = value
		end
	end
	
	-- add possessive target's caster's for event-specific data
	-- Stonarius' or Meneldill's with the proper apostrophe
	Substitutions["target's"] = self:MakePossessive( Substitutions.target )
	Substitutions["caster's"] = self:MakePossessive( Substitutions.caster )

	-- duplicate all the keys with localized variable names, if available
	for var,value in pairs(Substitutions) do
		local lkey = "<"..var..">"
		if L[ lkey ] then
			local newkey = string.lower( L[lkey] )
			if not Substitutions[newkey] then
				Substitutions[newkey] = Substitutions[var]
			end
		end
	end
	
	return Substitutions
end


function SpeakinSpell:DoSubstitution(var, Substitutions)
	-- do case insensitive lookup of variable names
	local varkey = string.lower(var)

	-- shuffle the random substitutions each time they are used within the same speech
	self:AddRandomSubstitutions(Substitutions)

	-- check for var = "<target|___>", or localized equivalent
	-- this should normally work the same as <target>
	-- however if the target is the player, then show ___ instead of the player's own name
	-- For example "Hey no hitting on <target|me>!" results in
	-- "Hey no hitting on Fred!" if cast on Fred
	-- "Hey no hitting on me!" if cast on yourself

	local ndxPipe = string.find(varkey, "|")
	if ndxPipe and (ndxPipe >= 0) then
		varkey = string.sub(varkey, 0, ndxPipe-1)
		local me = string.sub(var, ndxPipe+1) -- NOTE: use original Case Sensitive var name to get "me"
		
		-- KLUGE:
		-- behavior of string.sub seems inconsistent, sometimes ndxPipe+1 gets the | character, which then throws a LUA error
		-- but ndxPipe+2 sometimes misses the first letter of "me"
		-- so we use ndxPipe+1 and check for the leading |
		while string.find(me,"|") do
			me = string.sub(me, 2)
		end
		
		-- use "me" instead of my own name if targetting myself
		-- this applies to any and all substitution variables!
		if Substitutions[varkey] == UnitName("player") then
			return me
		end
		-- else, ignore the "me" part
		
		-- TODOFUTURE: consider using the logic above as a replacement for unknowns as well
		--		"if not Substitutions[varkey] return me end"
		--		would that need a different delimiter?
	end

	-- perform the substitution
	if	Substitutions[ varkey ] then
		return Substitutions[varkey]
	end

	-- its not a special substitution, so try for a simple UnitName API check
	local unit = UnitName(varkey)
	if unit then
		self:DebugMsg(nil,"Using UnitName("..varkey..")="..unit)
		return unit
	end
	unit = UnitName( self:MakeUnposessive(varkey) )
	if unit then
		self:DebugMsg(nil,"Using possessive form UnitName("..varkey..")="..unit)
		return self:MakePossessive(unit)
	end
	
	-- if we got this far then we did not find a value for this variable
	-- so return the variable unchanged, with wrapping < > restored
	
	-- NOTE: sending "|me" to SendChatAPI will throw a LUA error about an invalid escape code
	-- TODOFUTURE: figure out how to print a regular "|" ... tried "\|" and "||" but both failed
	--		this is also doing a weird thing duplicating the - character, i get "--" for every "|" ... why?
	return "<" .. string.gsub(var,"|","-") .. ">"
end


function SpeakinSpell:ProcessSubstitutionVars(msg, DetectedEvent)
	local Substitutions = self:BuildSubstitutionTable(DetectedEvent)
	return string.gsub(msg, "<(.-)>", 
		function(var) 
			return self:DoSubstitution(var, Substitutions) 
		end
	)
end



-- return true if this is a slash command
-- return false if this is other default text
function SpeakinSpell:ProcessChatSlashCommand(text, DetectedEvent)
	local funcname = "ProcessChatSlashCommand"
	
	if ( strsub(text, 1, 1) ~= "/" ) then
		return false
	end
	
	-- issue a warning if this a secure command like /cast
	-- If the string is in the format "/cmd blah", command will be "/cmd"
	local command = strmatch(text, "^(/[^%s]+)") or "";
	if IsSecureCmd(command) then
		self:Print(string.format(L["WARNING: can't execute protected command: %s"],text))
		return true
	end
	
	-- if this is a recursive call into a slash command, it needs some special handling
	-- do it with internal calls instead of passing through the global chat parser and back again
	if self:ProcessRecursiveUserMacro(text, DetectedEvent) then
		return true
	end
	
	self:DebugMsg(funcname,"executing slash command: " .. text)
	
	-- create a hidden edit box to parse and send the slash command
	if not SpeakinSpellTempEditBox then
		SpeakinSpellTempEditBox = CreateFrame("EditBox", "SpeakinSpellTempEditBox", UIParent)
		SpeakinSpellTempEditBox:Hide()
	end
	
	-- send the command
	SpeakinSpellTempEditBox:SetText(text) 
	ChatEdit_SendText(SpeakinSpellTempEditBox)

	return true
end



function SpeakinSpell:SplitLines(msg)
	local s1 = string.gsub(msg, "<newline>", "\n")
	local s2 = string.gsub(s1, L["<newline>"], "\n")
	local t = { strsplit("\n",s2) }
	return t
end


function SpeakinSpell:SayOneLine(line, channel, WhisperTarget, DetectedEvent)
	-- if this message is a slash command, send it through an editBox function to parse and execute the command
	--NOTE: a specified slash command overrides the default channel and whisper features below
	--		it doesn't make sense to combine them
	if self:ProcessChatSlashCommand(line, DetectedEvent) then
		return
	end

	-- say the message to the chat, if applicable
	if channel then
		if channel == "SPEAKINSPELL CHANNEL" then
			self:Print(line)
		else
			local language = nil --TODOFUTURE: allow for other languages
			--TODOFUTURE: should this check for "|" for invalid escape sequences to prevent LUA errors? or allow color changes and stuff?
			SendChatMessage(line, channel, language);
		end
	end
	
	-- send the message to a whisper to the target, if applicable
	if WhisperTarget then
		-- NOTE: always use Common(nil) for the language just to ensure the person you whisper can read what you say
		SendChatMessage(line, "WHISPER", nil, DetectedEvent.target);
	end
end


function SpeakinSpell:SpeakForSpell(DetectedEvent)
	local funcname = "SpeakForSpell"
	
	-- determine which channel we will speak in
	channel = self:GetChatChannelForSpell(DetectedEvent.EventTableEntry)
	
	--TODOFUTURE: make it an option to whisper yourself - but be wary of the auto self-cast logic
	local WhisperTarget = DetectedEvent.EventTableEntry.WhisperTarget and DetectedEvent.target and (DetectedEvent.target ~= "") and (DetectedEvent.target ~= "Unknown") and (DetectedEvent.target ~= UnitName("player")) and UnitIsFriend("player",DetectedEvent.target)

	if not channel and not WhisperTarget then
		-- this spell's speeches have no target for the current situation
		-- NOTE: this rule applies even if a slash command override is present in the randomly selected message text
		self:DebugMsg(funcname,"current scenario has no message targets for " .. DetectedEvent.SpellLink)
		return
	end
	
	local msg = self:SelectRandomMessage(DetectedEvent)
	
	-- record the time when we spoke for this spell, along with the last target, and last message spoken
	-- to support duplicate prevent, announcement limits, and cooldowns
	-- NOTE: record the unprocessed message text, so that changing the value of substitution variables
	--		does not make us think it's a new random message selection
	self:SaveAnnouncementHistory(DetectedEvent, msg)
	
	-- process substitution variables
	msg = self:ProcessSubstitutionVars(msg, DetectedEvent)

	-- break up the message into separate lines, if applicable
	local lines = self:SplitLines(msg)

	for _,line in ipairs(lines) do
		self:SayOneLine(line, channel, WhisperTarget, DetectedEvent)
	end
end

